from helpers import Simulator, Prior
import argparse
import os
from pathlib import Path
from lampe.data import H5Dataset, JointLoader

def main(method, size, discretisation, error, use_mtot_str):
    
    if use_mtot_str=="False":
        use_mtot=False
    elif use_mtot_str=="True":
        use_mtot=True
    else:
        raise ValueError("--use_mtot must be True or False")
    
    LOWER = [4.0, 1e-5, 180.0, 0.0, 0.0, 0.0, 50.0, 0.2]
    UPPER = [100.0, 0.99, 0.0, 360.0, 360.0, 1.0, 55.0, 3.0]

    prior = Prior(LOWER, UPPER)
    simulator = Simulator(method, 
                          scale=2e4,#3e3,#2e4, 
                          discretisation=discretisation, 
                          error=error, 
                          use_plx=False, 
                          use_mtot=use_mtot, 
                          prior=prior)

    loader = JointLoader(prior, simulator, batch_size=16, vectorized=True)
    
    scratch = os.environ.get('SCRATCH', '')
    path = Path(scratch) / 'dataset'
    path.mkdir(parents=True, exist_ok=True)

    H5Dataset.store(
        loader, 
        f'{path}/generic-{method}-disc_{discretisation}-err_{error}-mtot_{use_mtot}-test.h5', 
        size=(2**size)//64, # 130.000
        overwrite=True)
    H5Dataset.store(
        loader, 
        f'{path}/generic-{method}-disc_{discretisation}-err_{error}-mtot_{use_mtot}-val.h5', 
        size=(2**size)//8, # 1.000.000
        overwrite=True)
    H5Dataset.store(
        loader, 
        f'{path}/generic-{method}-disc_{discretisation}-err_{error}-mtot_{use_mtot}-train.h5', 
        size=2**size, # 8.400.000
        overwrite=True)

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="Generate betapic datasets")
    parser.add_argument("--method", type=str, default="Deepset", help="The method to use for the simulator")
    parser.add_argument("--size", type=int, default=23, help="The exponent of 2 for the training dataset size")
    parser.add_argument("--discretisation", type=int, default=180, help="Discretisation of the time")
    parser.add_argument("--error", type=int, default=10, help="Error on the observations")
    parser.add_argument("--use_mtot", type=str, default="False", help="Use mtot as input or not")
    args = parser.parse_args()
    main(args.method, args.size, args.discretisation, args.error, args.use_mtot)