import numpy as np
import orbitize
from orbitize import read_input, system, priors, sampler
import matplotlib.pyplot as plt

def main():
    data_table = read_input.read_file('HR8799e.csv')
    
    # change 0 stdev with 1s
    data_table["quant1_err"][data_table["quant1_err"] == 0] = 1 
    data_table["quant2_err"][data_table["quant2_err"] == 0] = 1
    
    num_planets = 1
    total_mass = 1.43 # [Msol]
    plx = 24.217 # [mas]
    mass_err = 0.0 # [Msol]
    plx_err = 0.0 # [mas]
    

    sys = system.System(
        num_planets, data_table, total_mass,
        plx, mass_err=mass_err, plx_err=plx_err,
        tau_ref_epoch=50000
    )

    lab = sys.param_idx

    sys.sys_priors[lab['sma1']] = priors.UniformPrior(4.0, 100.0)
    sys.sys_priors[lab['ecc1']] = priors.UniformPrior(0.00001, 0.99)
    sys.sys_priors[lab['inc1']] = priors.UniformPrior(0, np.pi)
    sys.sys_priors[lab['aop1']] = priors.UniformPrior(0, 2*np.pi)
    sys.sys_priors[lab['pan1']] = priors.UniformPrior(0, 2*np.pi)
    sys.sys_priors[lab['tau1']] = priors.UniformPrior(0, 1)

        
    # number of temperatures & walkers for MCMC
    num_temps = 20
    num_walkers = 1000

    # number of steps to take
    n_orbs = 10000000

    mcmc_sampler = sampler.MCMC(sys, num_temps, num_walkers, num_threads=10)
    
    sma_cen = 15.0
    sma_sig = 1.0
    ecc_cen = 0.16
    ecc_sig = 0.02
    inc_cen = 0.436332 
    inc_sig = 0.0272665  

    walker_centres = np.array(
        [sma_cen, ecc_cen, inc_cen]
    )
    walker_1sigmas = np.array(
        [sma_sig, ecc_sig, inc_sig]
    )

    curr_pos_shape = mcmc_sampler.curr_pos[:,:,:3].shape  # Get shape of walker positions
    new_pos = np.random.standard_normal(curr_pos_shape) * walker_1sigmas + walker_centres

    previous_last_three = mcmc_sampler.curr_pos[:,:,3:]

    complete_new_pos = np.concatenate((new_pos, previous_last_three), axis=2)

    mcmc_sampler.curr_pos = np.copy(complete_new_pos)

    _ = mcmc_sampler.run_sampler(n_orbs, output_filename='mcmc_HR8799e.hdf5')

if __name__ == '__main__':
    main()