from helpers import Prior, train 
from lampe.data import H5Dataset
from orbitize import DATADIR, read_input
import torch.nn as nn
import zuko 
import argparse
import wandb
import os
from pathlib import Path

def main(method, size, discretisation, error, use_mtot_str):
    if use_mtot_str=="False":
        use_mtot=False
    elif use_mtot_str=="True":
        use_mtot=True
    else:
        raise ValueError("--use_mtot must be True or False")
        
    scratch = os.environ.get('SCRATCH', '')
    path = Path(scratch) / 'dataset'
    path.mkdir(parents=True, exist_ok=True)
    
    trainset = H5Dataset(f'{path}/generic-{method}-disc_{discretisation}-err_{error}-mtot_{use_mtot}-train.h5', batch_size=2048, shuffle=True)  
    validset = H5Dataset(f'{path}/generic-{method}-disc_{discretisation}-err_{error}-mtot_{use_mtot}-val.h5', batch_size=2048, shuffle=True)
    
    LOWER = [4.0, 1e-5, 180.0, 0.0, 0.0, 0.0, 50.0, 0.2]
    UPPER = [100.0, 0.99, 0.0, 360.0, 360.0, 1.0, 55.0, 3.0]

    priors = Prior(LOWER, UPPER) # Needed to post-process the data in the training phase

    num_obs = discretisation 
    
    wandb.login()
    train(
        trainset=trainset,
        validset=validset,
        prior=priors,
        epochs=512,#1024,
        NPE_hidden_features=[256] * 3,
        flow = zuko.flows.spline.NSF,
        transforms = 3,
        num_obs=num_obs,
        mass_as_input=use_mtot,
        use_wandb=True,
        embedding=method
    )   

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Train model")
    parser.add_argument("--method", type=str, default="Deepset", help="The method to use for the simulator")
    parser.add_argument("--size", type=int, default=23, help="The exponent of 2 for the training dataset size")
    parser.add_argument("--discretisation", type=int, default=180, help="Discretisation of the time")
    parser.add_argument("--error", type=int, default=10, help="Error on the observations")
    parser.add_argument("--use_mtot", type=str, default="False", help="Use mtot as input or not")
    args = parser.parse_args()
    main(args.method, args.size, args.discretisation, args.error, args.use_mtot)